﻿using EPiServer.Commerce.Order;
using EPiServer.Framework.Localization;
using EPiServer.Security;
using EPiServer.ServiceLocation;
using Mediachase.Commerce.Core.Features;
using Mediachase.Commerce.Customers;
using Mediachase.Commerce.Extensions;
using Mediachase.Commerce.Orders;
using Mediachase.Commerce.Orders.Managers;
using Mediachase.Commerce.Plugins.Payment;
using Mediachase.Commerce.Security;
using Mediachase.Data.Provider;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;

namespace EPiServer.Business.Commerce.Payment.DIBS
{
    public class DIBSPaymentGateway : AbstractPaymentGateway, IPaymentPlugin
    {
        private IFeatureSwitch _featureSwitch;
        private IInventoryProcessor _inventoryProcessor;
        private IOrderRepository _orderRepository;
        private LocalizationService _localizationService;

        private DIBSRequestHelper _dibsRequestHelper;

        private IOrderForm _orderForm;

        public DIBSPaymentGateway():this(
            ServiceLocator.Current.GetInstance<IFeatureSwitch>(),
            ServiceLocator.Current.GetInstance<IInventoryProcessor>(),
            ServiceLocator.Current.GetInstance<IOrderRepository>(),
            ServiceLocator.Current.GetInstance<LocalizationService>())
        {
        }

        public DIBSPaymentGateway(
            IFeatureSwitch featureSwitch,
            IInventoryProcessor inventoryProcessor,
            IOrderRepository orderRepository,
            LocalizationService localizationService)
        {
            _featureSwitch = featureSwitch;
            _inventoryProcessor = inventoryProcessor;
            _orderRepository = orderRepository;
            _localizationService = localizationService;

            _dibsRequestHelper = new DIBSRequestHelper();
        }

        public IOrderGroup OrderGroup { get; set; }

        /// <summary>
        /// Main entry point of ECF Payment Gateway.
        /// </summary>
        /// <param name="payment">The payment to process</param>
        /// <param name="message">The message.</param>
        /// <returns>return false and set the message will make the WorkFlow activity raise PaymentExcetion(message)</returns>
        public override bool ProcessPayment(Mediachase.Commerce.Orders.Payment payment, ref string message)
        {
            OrderGroup = payment.Parent.Parent;
            _orderForm = OrderGroup.Forms.FirstOrDefault(form => form.Payments.Contains(payment));
            return ProcessPayment(payment as IPayment, ref message);
        }

        public bool ProcessPayment(IPayment payment, ref string message)
        {
            if (HttpContext.Current == null)
            {
                return true;
            }

            if (payment == null)
            {
                message = Utilities.Translate("PaymentNotSpecified");
                return false;
            }

            if (OrderGroup == null)
            {
                message = Utilities.Translate("PaymentNotAssociatedOrderGroup");
                return false;
            }

            _orderForm = _orderForm ?? OrderGroup.Forms.FirstOrDefault(f => f.Payments.Contains(payment));
            if (_orderForm == null)
            {
                message = Utilities.Translate("PaymentNotAssociatedOrderForm");
                return false;
            }

            var purchaseOrder = OrderGroup as IPurchaseOrder;
            if (purchaseOrder != null)
            {
                if (payment.TransactionType == TransactionType.Capture.ToString())
                {
                    // return true meaning the capture request is done,
                    // actual capturing must be done on DIBS.
                    var result = _dibsRequestHelper.PostCaptureRequest(payment, purchaseOrder);
                    // result containing ACCEPTED means the request was successful
                    if (result.IndexOf("ACCEPTED") == -1)
                    {
                        message = "There was an error while capturing payment with DIBS";
                        return false;
                    }
                    return true;
                }

                if (payment.TransactionType == TransactionType.Credit.ToString())
                {
                    var transactionID = payment.TransactionID;
                    if (string.IsNullOrEmpty(transactionID) || transactionID.Equals("0"))
                    {
                        message = "TransactionID is not valid or the current payment method does not support this order type.";
                        return false;
                    }
                    // The transact must be captured before refunding
                    var result = _dibsRequestHelper.PostRefundRequest(payment, purchaseOrder);
                    if (result.IndexOf("ACCEPTED") == -1)
                    {
                        message = "There was an error while refunding with DIBS";
                        return false;
                    }
                    return true;
                }

                // right now we do not support processing the order which is created by Commerce Manager
                message = "The current payment method does not support this order type.";
                return false;
            }

            var cart = OrderGroup as ICart;
            if (cart != null && cart.OrderStatus == OrderStatus.Completed)
            {
                // return true because this shopping cart has been paid already on DIBS
                return true;
            }
            _orderRepository.Save(OrderGroup);

            var redirectUrl = Utilities.GetUrlFromStartPageReferenceProperty("DIBSPaymentPage");
            payment.Properties["NextAction"] = new Action(() =>
            {
                HttpContext.Current.Response.Redirect(redirectUrl);
            });

            // If you're using workflow, uncomment the line below
            // (payment.Properties["NextAction"] as Action).Invoke();

            return true;
        }

        /// <summary>
        /// Processes the unsuccessful transaction.
        /// </summary>
        /// <param name="cancelUrl">The cancel url.</param>
        /// <param name="errorMessage">The error message.</param>
        /// <returns>The url redirection after process.</returns>
        public string ProcessUnsuccessfulTransaction(string cancelUrl, string errorMessage)
        {
            return UriSupport.AddQueryString(cancelUrl, "message", errorMessage); ;
        }

        /// <summary>
        /// Processes the successful transaction, will be called when DIBS server processes 
        /// the payment successfully and redirect back.
        /// </summary>
        /// <param name="cart">The cart that was processed.</param>
        /// <param name="payment">The order payment.</param>
        /// <param name="transactionID">The transaction id.</param>
        /// <param name="orderNumber">The order number.</param>
        /// <param name="acceptUrl">The redirect url when finished.</param>
        /// <param name="cancelUrl">The redirect url when error happens.</param>
        /// <returns>The redirection url after processing.</returns>
        public string ProcessSuccessfulTransaction(ICart cart, IPayment payment, string transactionID, string orderNumber, string acceptUrl, string cancelUrl)
        {
            if (cart == null)
            {
                return cancelUrl;
            }

            var redirectionUrl = string.Empty;
            using (TransactionScope scope = new TransactionScope())
            {
                // Change status of payments to processed.
                // It must be done before execute workflow to ensure payments which should mark as processed.
                // To avoid get errors when executed workflow.
                PaymentStatusManager.ProcessPayment(payment);

                CheckoutCart(cart);

                // Save the transact from DIBS to payment.
                payment.TransactionID = transactionID;

                var purchaseOrder = MakePurchaseOrder(cart, orderNumber);

                // Commit changes
                scope.Complete();

                if (HttpContext.Current.Session != null)
                {
                    HttpContext.Current.Session.Remove("LastCouponCode");
                }

                redirectionUrl = UpdateAcceptUrl(purchaseOrder, payment, acceptUrl);
            }

            return redirectionUrl;
        }


        private IPurchaseOrder MakePurchaseOrder(ICart cart, string orderNumber)
        {
            // Save changes
            //this might cause problem when checkout using multiple shipping address because ECF workflow does not handle it. Modify the workflow instead of modify in this payment
            var purchaseOrderLink = _orderRepository.SaveAsPurchaseOrder(cart);
            var purchaseOrder = _orderRepository.Load<IPurchaseOrder>(purchaseOrderLink.OrderGroupId);

            UpdateOrder(purchaseOrder, orderNumber);

            UpdateLastOrderOfCurrentContact(CustomerContext.Current.CurrentContact, purchaseOrder.Created);
            
            AddNoteToPurchaseOrder($"New order placed by {PrincipalInfo.CurrentPrincipal.Identity.Name} in Public site", purchaseOrder);

            _orderRepository.Save(purchaseOrder);

            // Remove old cart
            _orderRepository.Delete(cart.OrderLink);

            return purchaseOrder;
        }

        private void UpdateOrder(IPurchaseOrder purchaseOrder, string orderNumber)
        {
            purchaseOrder.OrderStatus = OrderStatus.InProgress;
            purchaseOrder.OrderNumber = orderNumber;

            // Update display name of product by current language
            Utilities.UpdateDisplayNameWithCurrentLanguage(purchaseOrder);
        }

        /// <summary>
        /// Update last order time stamp which current user completed.
        /// </summary>
        /// <param name="contact">The customer contact.</param>
        /// <param name="datetime">The order time.</param>
        private void UpdateLastOrderOfCurrentContact(CustomerContact contact, DateTime datetime)
        {
            if (contact != null)
            {
                contact.LastOrder = datetime;
                contact.SaveChanges();
            }
        }

        private string UpdateAcceptUrl(IPurchaseOrder purchaseOrder, IPayment payment, string acceptUrl)
        {
            var redirectionUrl = UriSupport.AddQueryString(acceptUrl, "success", "true");
            redirectionUrl = UriSupport.AddQueryString(redirectionUrl, "contactId", purchaseOrder.CustomerId.ToString());
            redirectionUrl = UriSupport.AddQueryString(redirectionUrl, "orderNumber", purchaseOrder.OrderLink.OrderGroupId.ToString());
            redirectionUrl = UriSupport.AddQueryString(redirectionUrl, "notificationMessage", string.Format(_localizationService.GetString("/OrderConfirmationMail/ErrorMessages/SmtpFailure"), payment.BillingAddress.Email));
            redirectionUrl = UriSupport.AddQueryString(redirectionUrl, "email", payment.BillingAddress.Email);
            return redirectionUrl;
        }

        /// <summary>
        /// Validates and checks out a cart.
        /// </summary>
        /// <param name="cart">The cart</param>
        private void CheckoutCart(ICart cart)
        {
            if (_featureSwitch.IsSerializedCartsEnabled())
            {
                cart.AdjustInventoryOrRemoveLineItems((item, issue) => { }, _inventoryProcessor);
            }
            else
            {
                // Execute CheckOutWorkflow with parameter to ignore running process payment activity again.
                var isIgnoreProcessPayment = new Dictionary<string, object> { { "PreventProcessPayment", true } };
                OrderGroupWorkflowManager.RunWorkflow((OrderGroup)cart, OrderGroupWorkflowManager.CartCheckOutWorkflowName, true, isIgnoreProcessPayment);
            }
        }

        /// <summary>
        /// Adds the note to purchase order.
        /// </summary>
        /// <param name="note">The note detail.</param>
        /// <param name="purchaseOrder">The purchase order.</param>
        private void AddNoteToPurchaseOrder(string note, IPurchaseOrder purchaseOrder)
        {
            var orderNote = purchaseOrder.CreateOrderNote();
            orderNote.Type = OrderNoteTypes.System.ToString();
            orderNote.CustomerId = PrincipalInfo.CurrentPrincipal.GetContactId();
            orderNote.Title = note.Substring(0, Math.Min(note.Length, 24)) + "...";
            orderNote.Detail = note;
            orderNote.Created = DateTime.UtcNow;
            purchaseOrder.Notes.Add(orderNote);
        }
    }
}